// -*- C++ -*-
//
// Copyright (C) 2004  Richard Guenther
//
// This file is part of FreePOOMA.
//
// FreePOOMA is free software; you can redistribute it and/or modify it
// under the terms of the Expat license.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the Expat
// license for more details.
//
// You should have received a copy of the Expat license along with
// FreePOOMA; see the file LICENSE.
//

#include <iostream>

#include "Pooma/Arrays.h"
#include "Evaluator/PatchFunction.h"
#include "Evaluator/PatchCode.h"

extern "C" {
	// prototypes can be constructed using f2c for example
	void fortran_compute_(int *, int *, double *, double *, double *);
}

// You still need to be careful specifying the right tag class to
// PatchFunction<> to ensure proper read/write requests are issued
// before computation.

struct MyFunction {
	MyFunction() {}

	typedef const Array<2, double, BrickView> Array_t;
	void apply(Array_t &f1, Array_t &f2, Array_t &f3) const
	{
		// As we come here with BrickView engines we can't do
		// more than supporting operations on the natural physical
		// domain of an array/field with the guard cell setup
		// hidden in the fortran source (it needs to know anyway).
		int ni = f1.domain()[0].size();
		int nj = f1.domain()[1].size();
		fortran_compute_(&ni, &nj,
				  f1.engine().dataBlock().beginPointer(),
				  f2.engine().dataBlock().beginPointer(),
				  f3.engine().dataBlock().beginPointer());
	}
};


// With PatchCode we need to supply appropriate information via scalarCodeInfo().
// The main reason for PatchCode is the ability to use guards and have them
// updated by the FreePOOMA machinery automatically, which PatchFunction does
// not support.

struct MyFunction2 {
  MyFunction2() {}

  template <class A1, class A2, class A3>
  void operator()(const A1& a, const A2& b, const A3 c,
		  const Interval<2>& domain) const
  {
    // We won't be able to make use of a custom evaluation domain here,
    // so don't bother to try to make that information available to the
    // fortran code.
    int ni = domain[0].size();
    int nj = domain[1].size();
    fortran_compute_(&ni, &nj,
		      a.engine().dataBlock().beginPointer(),
		      b.engine().dataBlock().beginPointer(),
		      c.engine().dataBlock().beginPointer());
  }

  void scalarCodeInfo(ScalarCodeInfo& i) const
  {
    i.arguments(3);
    i.dimensions(2);
    i.write(0, true);
    i.useGuards(0, false);
  }
};


int main(int argc, char **argv)
{
	Pooma::initialize(argc, argv);

	Interval<2> I(9, 8);
	Loc<2> grid(3, 4);

	GridLayout<2> layout(I, grid, GuardLayers<2>(1), DistributedTag());
	Array<2, double, MultiPatch<GridTag, Brick> > A(layout), B(layout), C(layout);


	// Use PatchFunction.

	A = 0.0;
	B(I) = iota(I).comp(0);
	C(I) = iota(I).comp(1);

	PatchFunction<MyFunction, PatchTag3>()(A(I), B(I), C(I));

	C = 0.0;
	C(I) = iota(I).comp(0) + iota(I).comp(1);
	bool res = all(A == C);
	if (res)
		std::cout << "All ok with PatchFunction." << std::endl;
	else
		std::cout << A << "doesn't match\n" << C << std::endl;


	// Use PatchCode

	A = 0.0;
	B(I) = iota(I).comp(0);
	C(I) = iota(I).comp(1);

	PatchCode<MyFunction2>()(A, I, B, C);

	C = 0.0;
	C(I) = iota(I).comp(0) + iota(I).comp(1);
	res = all(A == C);
	if (res)
		std::cout << "All ok with PatchCode." << std::endl;
	else
		std::cout << A << "doesn't match\n" << C << std::endl;


	Pooma::finalize();
	return 0;
}
